import os
from configparser import ConfigParser
from dataclasses import MISSING, fields
from typing import Any, Dict, List, Optional, Type, TypeVar

from magiconf.flags import load_flags
from magiconf.errors import ConfigError


T = TypeVar("T")


def load_env(fields: List[str], env_prefix: Optional[str] = None) -> Dict[str, str]:
    prefix, prefixed_fields = "", [*fields]
    if env_prefix is not None:
        prefix = f"{env_prefix.rstrip('_')}_"
        prefixed_fields = [(prefix + name).lower() for name in fields]

    return {
        # fmt: off
        k.lower()[len(prefix):]: v
        # fmt: on
        for k, v in os.environ.items()
        if k.lower() in prefixed_fields
    }


def load_config(
    fields: List[str], content: str, section: str = "default"
) -> Dict[str, str]:
    parser = ConfigParser()
    parser.read_string(content)
    # TODO: support sections
    return {k: v for k, v in parser.items(section) if k in fields}


def load(conf_cls: Type[T], env_prefix: Optional[str] = None):
    field_names = [f.name for f in fields(conf_cls)]
    sources: List[Dict[str, Any]] = [
        load_flags(fields(conf_cls)),
        load_env(field_names, env_prefix),
    ]

    try:
        with open("config.ini", "r") as f:
            content = f.read()
            sources.append(load_config(field_names, content))
    except IOError:
        pass

    return _load(conf_cls, sources)


def _load(conf_cls: Type[T], sources: List[Dict[str, Any]]) -> T:
    init_kw = {}
    for field in fields(conf_cls):
        for src in sources:
            if field.name in src:
                init_kw[field.name] = src[field.name]
                break
        else:
            is_requried = (
                getattr(field, "default", None) == MISSING
                and getattr(field, "default_factory", None) == MISSING
            )
            if is_requried:
                raise ConfigError(f"{field.name} is required but is missing")

        if field.name in init_kw and not isinstance(init_kw[field.name], field.type):
            raise TypeError(
                "{} is of wrong type ({}, expected {})".format(
                    field.name, type(init_kw[field.name]), field.type
                )
            )

    return conf_cls(**init_kw)  # type: ignore



from argparse import ArgumentParser
from dataclasses import Field
from typing import Dict, Iterable

from magiconf.errors import ConfigError


def boolean(val):
    if val is None or val == "":
        return True

    if val in ("true", "True"):
        return True

    if val in ("false", "False"):
        return False

    raise ConfigError(f"{val} is not a boolean value")


class ArgumentConfigParser(ArgumentParser):
    def error(self, message):
        raise ConfigError(message)


def load_flags(fields: Iterable[Field]) -> Dict[str, str]:
    p = ArgumentConfigParser()
    for f in fields:
        # Use append to support specifying argument multiple times
        kwargs = {"action": "append", "type": f.type, "nargs": "?"}
        if f.type == bool:
            # We don't use store_true as it sets up False as default value
            # which might me unwanted, e.g. when default value in class is set
            # to True
            kwargs = {
                "action": "append",
                "type": boolean,
                "nargs": "?",
                "const": "",
            }

            # Add --no- version of the flag to support setting falsy values
            p.add_argument(
                f"--no-{f.name}", action="store_const", const=False, dest=f.name
            )

        p.add_argument(f"--{f.name}", **kwargs)  # type: ignore

    parsed, _ = p.parse_known_args()

    result = {}
    for k, v in parsed.__dict__.items():
        if v is None:
            continue

        if isinstance(v, list):
            if len(v) > 1:
                raise ConfigError(f"--{k} specified multiple times")

            if len(v) != 0:
                result[k] = v[0]
        else:
            result[k] = v

    return result



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['magiconf']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'magiconf',
    'version': '0.1.2',
    'description': 'A small library to automagically load and parse configuration from either environment, configuration file or command line arguments.',
    'long_description': None,
    'author': 'Vladimir Kotikov',
    'author_email': 'kotikov.vladimir@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
